/* htable.c -- Brief hash lookup implementation
 * SEE htable.h for details
 */

#include <string.h>
#include <stdlib.h>
#include "htable.h"
#include "stddefs.h"

/* -- hash system ADO -- */

typedef struct {
	char	key[MAX_KEY_SIZE];
	hEntry	entry;
} htableEntryTag;

struct htableTag {
	unsigned	numEntries;
	unsigned	hashSize;
	htableEntryTag	*table;
} hash_ADO;



/* -- Non-exported hash functons' defintions -- */

unsigned hash(char *text);
htableEntryTag *hashEntry(unsigned index);


/* -- Hash system functions    -- */
/* -- See htable.h for details -- */

/* -- Function hashInitTable */

bool hashInitTable(unsigned hashSize)
{
	int		index;
	htableEntryTag	*record;

	hash_ADO.numEntries = 0;
	hash_ADO.hashSize = hashSize;

	hash_ADO.table = (htableEntryTag *) malloc(hashSize * MAX_KEY_SIZE * sizeof(htableEntryTag));

	if (hash_ADO.table == NULL)
		return FALSE;				/* Couldn't allocate table */

	for (index = 0; index < hashSize; index++) {
		strcpy(hashEntry(index) -> key, HASH_EMPTY);
	}

	return TRUE;
}


/* -- Function hashDestroyTable -- */

bool hashDestroyTable(void)
{
   free(hash_ADO.table);
	return TRUE;
}


/* -- Function hashInsert */

bool hashInsert(char *key, hEntry entry)
{
	unsigned	hashed, insert;
	unsigned	hashSize;
	bool		contSearch;
	htableEntryTag	*record;

	if (!hashIsValidKey(key))
		return FALSE;				/* Failed precondition */

	if (hashIsIn(key))
		 return FALSE; 				/* Failed precondition */

	if (hashIsFull())				/* Failed precondition */
		return FALSE;

	insert = hashed = hash(key);
	hashSize = hash_ADO.hashSize;
	contSearch = TRUE;

	while (contSearch) {
		/* Insert if either empty or deleted */
		if (	(strcmp(hashEntry(insert) -> key, HASH_EMPTY) == 0) ||
				(strcmp(hashEntry(insert) -> key, HASH_DELETED) == 0)) {
			record = hashEntry(insert);
			strcpy(record -> key, key);
			record -> entry = entry;
			hash_ADO.numEntries++;
			return TRUE;
		}

		insert += 1;				/* Search in next position */

		if (insert >= hashSize)
			insert = 0;			/* Wrap around table */

		if (insert == hashed) {
			contSearch = FALSE;		/* Searched entire table */
         break;
		}

	}

	return FALSE;					 /* Shouldn't reach */
}


/* -- Function hashDelete */

bool hashDelete(char *key)
{
	unsigned	hashed, insert;
	unsigned	hashSize;
	bool		contSearch;

	if (!hashIsValidKey(key))
		return FALSE;				/* Failed precondition */

	if (!hashIsIn(key))
		return FALSE;				/* Failed precondition */

	if (hashIsEmpty())				/* Failed precondition */
   	return FALSE;

	insert = hashed = hash(key);
	hashSize = hash_ADO.hashSize;
	contSearch = TRUE;

	while (contSearch) {
		/* Delete if matches */
		if (strcmp(hashEntry(insert) -> key, key) == 0) {
			strcpy(hashEntry(insert) -> key, HASH_DELETED);
			hash_ADO.numEntries--;
			return TRUE;
		}

		insert += 1;				/* Search in next position */

		if (insert >= hashSize)
			insert = 0;			/* Wrap around table */

		if (insert == hashed) {
			contSearch = FALSE;		/* Searched entire table */
			break;
		}
	}

	return FALSE;					/* Should never get here */
}


/* -- Function hashLoadFactor */

unsigned hashLoadFactor(void)
{
	if (hashIsEmpty())
		return 0;

	else
		return (unsigned) ((hash_ADO.numEntries * 100) / hash_ADO.hashSize);
}


/* -- Function hashNumEntries */

unsigned hashNumEntries(void)
{
	return hash_ADO.numEntries;
}


/* -- Function hashIsIn */

bool hashIsIn(char *key)
{
	unsigned hashed, insert;
	unsigned	hashSize;
	bool		contSearch;

	if (!hashIsValidKey(key))
		return FALSE;				/* Failed precondition */

	insert = hashed = hash(key);			/* Find initial hash */
	hashSize = hash_ADO.hashSize;
	contSearch = TRUE;				/* Do search while contSearch */

	while (contSearch) {
		if (strcmp(hashEntry(insert) -> key, key) == 0) return TRUE;	/* Found key */
		if (strcmp(hashEntry(insert) -> key, HASH_EMPTY) == 0) {		/* never inserted */
			contSearch = FALSE;
			break;
		}

		insert += 1;						/* Search in next position */

		if (insert >= hashSize)
			insert = 0;						/* Wrap around table */

		if (insert == hashed) {
			contSearch = FALSE;			/* Searched entire table */
			break;
		}
	}

	return FALSE;
}

/* -- Function hashValue */

hEntry	hashValue(char *key)
{
	unsigned	hashed, index;
	unsigned	hashSize;
	bool		contSearch, found;

	if (!hashIsValidKey(key))
		return 0;				/* Failed precondition */

	if (!hashIsIn(key))
		return 0;				/* Failed precondition */

	index = hashed = hash(key);			/* Find initial hash */
	hashSize = hash_ADO.hashSize;
	contSearch = TRUE;				/* Do search while contSearch */
	found = FALSE;

	while (contSearch) {
		if (strcmp(hashEntry(index) -> key, key) == 0) {
			contSearch = FALSE;		/* Found key */
			found = TRUE;
			break;
		}
		if (strcmp(hashEntry(index) -> key, HASH_EMPTY) == 0) {		/* never inserted */
			contSearch = FALSE;
			break;
		}

		index += 1;						/* Search in next position */

		if (index >= hashSize)
			index = 0;						/* Wrap around table */

		if (index == hashed) {
			contSearch = FALSE;			/* Searched entire table */
			break;
		}
	}
	if (!found)
		return 0;
	else
		return hashEntry(index) -> entry;
}


/* -- Function hashIsFull */

bool hashIsFull(void)
{
	if (hash_ADO.numEntries == hash_ADO.hashSize)
		return TRUE;

	else
		return FALSE;
}


/* -- Function hashIsFull */

bool hashIsEmpty(void)
{
	if (hash_ADO.numEntries == 0)
		return TRUE;

	else
		return FALSE;
}


/* -- Function hashIsValidKey */

bool hashIsValidKey(char *key)
{
	if (strlen(key) > MAX_KEY_SIZE)
		return FALSE;

	else
		return TRUE;
}


/* -- Non-exported hash functions -- */

/* -- Function hash
 * Pre: 'text' is a valid nul-terminated string
 * Post: returned has value of string
 */
unsigned hash(char *text)
{
	unsigned	hash;

	hash = 0;
	while (*text != EOS) {
		hash += *text;
		text++;
	}

	hash %= hash_ADO.hashSize;

	return hash;
}

/* -- Function hashEntry
 * Pre: 'index' is a valid hash
 * Post: (a pointer to the indexed entry was returned) ||
 *			(an error occurred && NULL was returned)
 */
htableEntryTag *hashEntry(unsigned index)
{
	if (index >= hash_ADO.hashSize)
   	return NULL;

	return (hash_ADO.table + index * MAX_KEY_SIZE);
}


/* --- END of htable.c --- */

